/*
**  OSSP ex - Exception Handling (modified to fit into SimGrid)
**  Copyright (c) 2005 Martin Quinson
**  Copyright (c) 2002-2004 Ralf S. Engelschall <rse@engelschall.com>
**  Copyright (c) 2002-2004 The OSSP Project <http://www.ossp.org/>
**  Copyright (c) 2002-2004 Cable & Wireless <http://www.cw.com/>
**
**  This file is part of OSSP ex, an exception handling library
**  which can be found at http://www.ossp.org/pkg/lib/ex/.
**
**  Permission to use, copy, modify, and distribute this software for
**  any purpose with or without fee is hereby granted, provided that
**  the above copyright notice and this permission notice appear in all
**  copies.
**
**  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
**  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
**  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
**  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
**  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
**  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
**  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
**  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
**  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
**  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
**  SUCH DAMAGE.
**
**  ex.c: exception handling (compiler part)
*/

#include <stdio.h>
#include <stdlib.h>

#include "portable.h" /* execinfo when available */
#include "xbt/ex.h"

/* default __ex_ctx callback function */
ex_ctx_t *__xbt_ex_ctx_default(void) {
    static ex_ctx_t ctx = XBT_CTX_INITIALIZER;

    return &ctx;
}

/* default __ex_terminate callback function */
void __xbt_ex_terminate_default(xbt_ex_t *e)  {

  fprintf(stderr,
	  "** SimGrid: UNCAUGHT EXCEPTION: category: %s; value: %d\n"
	  "** %s\n"
	  "** Thrown by %s%s%s at %s:%d:%s\n",
	  xbt_ex_catname(e->category), e->value, e->msg,
	  e->procname, (e->host?"@":""),(e->host?e->host:"localhost"),
	  e->file,e->line,e->func);

#ifdef HAVE_EXECINFO_H
 {
  char **strings;
  size_t i;

  fprintf(stderr,"** Backtrace:\n");
  strings = backtrace_symbols (e->bt, e->used);

  for (i = 0; i < e->used; i++)
     printf ("   %s\n", strings[i]);

  free (strings);
 }
#endif

  abort();
}

/* the externally visible API */
ex_ctx_cb_t  __xbt_ex_ctx       = &__xbt_ex_ctx_default;
ex_term_cb_t __xbt_ex_terminate = &__xbt_ex_terminate_default;

void xbt_ex_free(xbt_ex_t e) {
  free(e.msg);
  free(e.procname);
}

/** \brief returns a short name for the given exception category */
const char * xbt_ex_catname(xbt_errcat_t cat) {
  switch (cat) {
  case unknown_error:   return  "unknown_err";
  case arg_error:       return "invalid_arg";
  case mismatch_error:  return "mismatch";
  case not_found_error: return "not found";
  case system_error:    return "system_err";
  case network_error:   return "network_err";
  case timeout_error:   return "timeout";
  case thread_error:    return "thread_err";
  default:              return "INVALID_ERR";
  }
}

#ifndef HAVE_EXECINFO_H
/* dummy implementation. We won't use the result, but ex.h needs it to be defined */
int backtrace (void **__array, int __size) {
  return 0;
}

#endif
